from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass
import torch.nn as nn

class ModelBase(ConfigBasedClass,nn.Module):
    def __init__(self,args,subconfig_keys):
        ConfigBasedClass.__init__(self, args, subconfig_keys)
        nn.Module.__init__(self)
        self.resume_epoch = None

    def _set_device(self, device):
        self.device = device

    # Call this method to get the device on which the model is currently located
    def _get_device(self):
        return self.device

    def _set_resume_epoch(self, epoch):
        self.resume_epoch = epoch

    # Call this method to get the epoch from which the model resumed training or inference
    def _get_resume_epoch(self):
        return self.resume_epoch    


from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass
from file_golem import Config

class ConfigCopier(ConfigBasedClass):
    def __init__(self, args):
        super().__init__(args)
        self.new_config_name = args.copy_config


    def __call__(self):
        self.check_availability()
        print('Compiling config...')

        config = self.data_io.load_config(self.global_config_name)
        if 'parent' in config:
            del config['parent']
        
        config['compile_origin'] = self.global_config_name
        self.data_io.save_data(Config, data_args = {
            Config.CONFIG_NAME: self.new_config_name,
            Config.DATA: config})


    def check_availability(self):
        if self.data_io.is_file_present(Config, data_args = {
            Config.CONFIG_NAME: self.new_config_name}):
            raise Exception(f'Config file {self.new_config_name} already exists. Please delete it or choose a different name.')
        return


import time

class Timer:

    def start(self):
        self.start_time = time.time()

    def stop(self):
        if self.start_time is None:
            raise ValueError("Timer has not been started.")
        elapsed_time = time.time() - self.start_time
        self.start_time = None
        return elapsed_time

    def format_time(self, elapsed_time):
        hours, rem = divmod(elapsed_time, 3600)
        minutes, seconds = divmod(rem, 60)
        return f"{int(hours):02}:{int(minutes):02}:{seconds:05.2f}"


import torch
import itertools
from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass
from ml_golem.model_loading_logic.model_config_keywords import ModelConfigKeywords
from torch.utils.data import DataLoader
from ml_golem.base_classes.dataset_base import DatasetBase
from ml_golem.base_classes.loss_base import LossBase
            
class DataIterationBase(ConfigBasedClass):

    def _initialize_dataloader(self,args,subconfig_keys):
        config = self.data_io.fetch_subconfig(
            self.global_config_name,
            subconfig_keys=subconfig_keys)

        if ModelConfigKeywords.DATASET.value not in config:
            return None
            
        dataset = self.instantiate_config_based_class(
            args,
            self.global_config_name,
            subconfig_keys=subconfig_keys+[ModelConfigKeywords.DATASET.value],
            default_class=DatasetBase)

        dataloader_config= self.data_io.fetch_subconfig(
            self.global_config_name,
            subconfig_keys=subconfig_keys+[ModelConfigKeywords.DATALOADER.value])
        
        dataloader = DataLoader(dataset,
            batch_size=dataloader_config.get(ModelConfigKeywords.BATCH_SIZE.value, 1), 
            num_workers=dataloader_config.get(ModelConfigKeywords.NUM_WORKERS.value,0),
            collate_fn=dataset._get_custom_collate_fn(),
            shuffle=dataloader_config.get(ModelConfigKeywords.CAN_SHUFFLE.value, self._can_shuffle()), 
            pin_memory=True)
        return dataloader

    def _initialize_loss(self,args,subconfig_keys):
        return self.instantiate_config_based_class(
            args,
            self.global_config_name,
            subconfig_keys=subconfig_keys+[ModelConfigKeywords.LOSS.value],
            default_class=LossBase)
    
    def _initialize_optimizer(self):
        self.weight_decay = self.data_io.fetch_config_field(
            self.config,
            subconfig_keys=[ModelConfigKeywords.LOSS.value, ModelConfigKeywords.WEIGHT_DECAY.value],
            is_required=False,
            default_value=0.0
        )

        self.learning_rate = float(self.config[ModelConfigKeywords.LEARNING_RATE.value])


        # Combine parameters from all models
        combined_parameters = itertools.chain(*[model.parameters() for model in self.models])
        optimizer = torch.optim.Adam(
            combined_parameters, 
            lr=self.learning_rate,
            weight_decay=self.weight_decay)
        
        return optimizer

    
    def _can_shuffle(self):
        return False



from ml_golem.overleaf.overleaf_datatypes import ProjectDatatype, OverleafDatatype,OverleafDirectory
from file_golem import SpecialDataArgs
from ml_golem.base_classes.data_io_object import DataIOObject

class OverleafTransfer(DataIOObject):
    def __init__(self,args):
        super().__init__(args)
        self.overleaf_subdir = self.data_io.system_config.get(SpecialDataArgs.OVERLEAF_SUBDIR.value)
        if SpecialDataArgs.OVERLEAF_MAPPINGS.value not in self.data_io.system_config:
            raise Exception(f'No dict named {SpecialDataArgs.OVERLEAF_MAPPINGS.value} found in the system config')
        self.mappings = self.data_io.system_config.get(SpecialDataArgs.OVERLEAF_MAPPINGS.value)

    def __call__(self):


        overleaf_dir_path = self.data_io.get_data_path(OverleafDirectory, data_args={
            OverleafDirectory.OVERLEAF_SUBDIR: self.overleaf_subdir
        })
        git_pull_command = (
            f'cd {overleaf_dir_path} && '
            'git pull'
        )
        self.data_io.run_system_command(git_pull_command)

        self.transfer_data()

        git_push_command = (
            f'cd {overleaf_dir_path} && '
            'git add . && '
            'git commit -m \"updating progress\" && '
            'git push'
        )
        self.data_io.run_system_command(git_push_command)


    def transfer_data(self):
        for entry in self.mappings:
            project_base, overleaf_base = next(iter(entry.items()))
            self.data_io.transfer_data(
                ProjectDatatype,
                OverleafDatatype,
                source_data_args = {
                    ProjectDatatype.SYSTEM_CONFIG_SPECIFIED_PATH: project_base
                },
                target_data_args = {
                    OverleafDatatype.OVERLEAF_SUBDIR: self.overleaf_subdir,
                    OverleafDatatype.SYSTEM_CONFIG_SPECIFIED_PATH: overleaf_base
                })


import torch
from ml_golem.base_classes.model_io_base import ModelIOBase

class ModelInference(ModelIOBase):
    def __init__(self,args,subconfig_keys):
        super().__init__(args,subconfig_keys)

        self._prepare_accelerator()

    def __call__(self):
        self._call_core_function_with_autocast(self._inference_core)


    def _inference_core(self):
        with torch.no_grad():  # Disable gradient computation
            if self.dataloader is None:
                results = self.make_forward_pass()
                self.save_inference_results(results)
            else:
                for input_data in self.dataloader:
                    results = self.make_forward_pass(input_data)
                    self.save_inference_results(results,input_data)

            self.complete_inference()

    
    def save_inference_results(self,output,input_data=None):
        model = self._get_callable_model(-1)
        if hasattr(model, 'save_results'):
            model.save_results(output, input_data)


    def complete_inference(self):
        model = self._get_callable_model(-1)
        if hasattr(model, 'complete_inference'):
            model.complete_inference()



from file_golem import AbstractDatatype, FileDatatypes, FilePathEntries
import os



class PathSeperatorDatatype(AbstractDatatype):
    FILE_DATATYPE = FileDatatypes.EMPTY
    SYSTEM_CONFIG_SPECIFIED_PATH = 'system_config_specified_path'

    @staticmethod
    def _split_path(data_args):
        base_path = data_args[PathSeperatorDatatype.SYSTEM_CONFIG_SPECIFIED_PATH]
        return base_path.split(os.sep)

class ProjectDatatype(PathSeperatorDatatype):
    RELATIVE_PATH_TUPLE = AbstractDatatype.RELATIVE_PATH_TUPLE + (
        {FilePathEntries.CUSTOM_LOGIC: '_split_path'},
    )


class OverleafDirectory(PathSeperatorDatatype):
    OVERLEAF_SUBDIR = 'overleaf_subdir'

    RELATIVE_PATH_TUPLE = AbstractDatatype.RELATIVE_PATH_TUPLE + (
        {FilePathEntries.CUSTOM_LOGIC: '_split_overleaf_path'},
    )

    @staticmethod
    def _split_overleaf_path(data_args):
        overleaf_subdir = data_args[OverleafDirectory.OVERLEAF_SUBDIR]
        return overleaf_subdir.split(os.sep)


class OverleafDatatype(OverleafDirectory):
    OVERLEAF_SUBDIR = 'overleaf_subdir'

    RELATIVE_PATH_TUPLE = OverleafDirectory.RELATIVE_PATH_TUPLE + (
        {FilePathEntries.CUSTOM_LOGIC: '_split_path'},
    )



import time
import io
import os
from enum import Enum
from itertools import product
from omegaconf import OmegaConf
from file_golem import Config, FilePathEntries
from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass
from ml_golem.datatypes import GridShellScript, GridSlurmScript, AbstractSlurmOutput, SlurmOutputStd, SlurmOutputErr
from ml_golem.model_loading_logic.model_config_keywords import ModelConfigKeywords
from ml_golem.main_loop_logic.reserved_keywords import ReservedArgKeywords
from ml_golem.base_classes.data_io_object import DataIOObject
from ml_golem.main_loop_logic.main_loop_messages import MainLoopMessages


class GridJobExecutionStyle(Enum):
    SEQUENTIAL = 'sequential'
    SLURM = 'slurm'
    NONE = 'none'

    

class GridJobBuilder(ConfigBasedClass):
    def __init__(self,args):
        super().__init__(args,subconfig_keys=[ModelConfigKeywords.GRID_JOB.value])

        self.grid_job_readable_labels = self.config.get(ModelConfigKeywords.GRID_JOB_READABLE_LABELS.value,True)

        self.has_grid_debug = self.config.get(ModelConfigKeywords.GRID_DEBUG.value, False)
        self.grid_actions = self.config.get(ModelConfigKeywords.GRID_ACTIONS.value)
        if len(self.grid_actions) == 0:
            raise Exception('No grid actions provided')
        self.grid_job_style = self.config.get(ModelConfigKeywords.GRID_JOB_STYLE.value)

        if self.grid_job_style == GridJobExecutionStyle.SLURM.value:
            self.slurm_config = self.data_io.fetch_subconfig(
                self.config,
                subconfig_keys=[ModelConfigKeywords.GRID_SLURM_CONFIG.value])

        grid_job_params = self.config.get(ModelConfigKeywords.GRID_JOB_PARAMS.value,[])
        self.has_grid_job_params = (len(grid_job_params) != 0)
        if self.has_grid_job_params:
            #Should be of format [[grid_key_a, grid_key_b, ...], [grid_key_c, grid_key_d, ...], ...]
            self.all_joined_grid_job_params = {}
            grid_job_joins = self.config.get(ModelConfigKeywords.GRID_JOB_JOINS.value, [])
            join_key_set = set()
            for i in range(len(grid_job_joins)):
                join_keys = sorted(grid_job_joins[i])
                common_length = None
                for join_key in join_keys:
                    if join_key not in grid_job_params.keys():
                        raise ValueError(f'Grid job join key {join_key} not found in grid job parameters {grid_job_params.keys()}')
                    if join_key in join_key_set:
                        raise ValueError(f'Grid job join key {join_key} is duplicated in grid job joins {grid_job_joins}')
                    join_key_set.add(join_key)

                    if common_length is None:
                        common_length = len(grid_job_params[join_key])
                    elif common_length != len(grid_job_params[join_key]):
                        raise ValueError(f'Grid job join key {join_key} has a different length than the other join keys in {grid_job_joins[i]}')
                joined_params = [[self._parse_grid_job_param(grid_job_params[join_key])[j] for join_key in join_keys] for j in range(common_length)]
                self.all_joined_grid_job_params[tuple(join_keys)] = joined_params
                
            singlton_params = list(set(grid_job_params.keys() - join_key_set))
            for singleton in singlton_params:
                parsed_param = self._parse_grid_job_param(grid_job_params[singleton])
                self.all_joined_grid_job_params[tuple([singleton])] = [[p] for p in parsed_param]
            self.config_list = []


    def _parse_grid_job_param(self,param):
        if isinstance(param, str):
            if ('-' in param) and all(part.isdigit() for part in param.split('-')) and (len(param.split('-')) == 2):
                start, end = map(int, param.split('-'))
                parsed_param = list(range(start, end + 1))
            else:
                raise ValueError(f"Invalid grid job parameter format: {param}. Currently, we only support lists or ranges in the format 'start-end'.")
        else:
            parsed_param = param
        return parsed_param


    def __call__(self):
        print('Building grid search...')

        self.build_configs()

        if len(self.grid_actions) == 0:
            raise Exception('No grid actions provided')
        if self.grid_job_style == GridJobExecutionStyle.SEQUENTIAL.value:
            if not self.has_grid_job_params:
                raise Exception('Grid job style is sequential, but no grid job parameters provided. Therefore it is recommended to run the action directly in the terminal instead of using the grid search builder.')
            self.build_and_execute_sequential_shell_script()
        elif self.grid_job_style == GridJobExecutionStyle.SLURM.value:
            self.build_and_execute_slurm_script()
        else:
            raise Exception(f'Grid search style {self.grid_job_style} not recognized')

    def build_and_execute_slurm_script(self):
        slurm_script_file_name, total_array_length, log_dir_name = self.build_slurm_script()        
        array_args = f'-a 0-{total_array_length-1} ' if self.has_grid_job_params else ''
        slurm_command = f'sbatch {array_args}./{slurm_script_file_name}'

        slurm_server_output = self.ssh_to_slurm_and_execute(slurm_command)
        print(f'You can check the slurm logs in the directory: {log_dir_name}')
        self.slurm_job_id = int(slurm_server_output[0].split()[-1])  # Assuming the job number is the last part of the output


    def is_slurm_job_complete(self):
        slurm_command = f'squeue --job {self.slurm_job_id}'
        slurm_server_output = self.ssh_to_slurm_and_execute(slurm_command)
        if len(slurm_server_output) <= 1:
            return True
        else:
            return False

            
    def ssh_to_slurm_and_execute(self, slurm_server_command):
        slurm_command_buffer = io.StringIO()
        slurm_server_credentials, slurm_project_directory = self.data_io.get_slurm_server_credentials_and_project_directory()
        
        ssh_to_execute_command = (slurm_server_credentials is not None) and (slurm_project_directory is not None)
        if not ssh_to_execute_command:
            print(f'Warning: slurm server credentials or slurm project directory not provided. Will try calling slurm locally.')
        
        if ssh_to_execute_command:
            slurm_command_buffer.write(f'ssh {slurm_server_credentials} \" cd {slurm_project_directory} && ')
        slurm_command_buffer.write(slurm_server_command)
        if ssh_to_execute_command:
            slurm_command_buffer.write(f' && exit \"')
        command = slurm_command_buffer.getvalue()
        slurm_command_buffer.close()
        slurm_server_output, slurm_exit_code = self.data_io.run_system_command(command,return_output=True, return_exit_code=True )
        if slurm_exit_code != 0:
            raise Exception(f'SLURM command failed with exit code {slurm_exit_code}. Please check the console output to debug')
        return slurm_server_output


    def build_slurm_script(self):
        script_buffer = io.StringIO()
        script_buffer.write(f'#!/bin/bash\n')
        
        partition = self.slurm_config.get('partition', 'a100')
        script_buffer.write(f'#SBATCH --partition={partition}\n')

        gpu = self.slurm_config.get('gpu', 0)
        script_buffer.write(f'#SBATCH --gres=gpu:{gpu}\n')

        cpu = self.slurm_config.get('cpu',1)
        script_buffer.write(f'#SBATCH -c {cpu}\n')

        time = self.slurm_config.get('time','01:00:00')
        script_buffer.write(f'#SBATCH -t {time}\n')

        memory = self.slurm_config.get('memory', '1G')
        script_buffer.write(f'#SBATCH --mem={memory}\n')

        data_args = {
            GridSlurmScript.CONFIG_NAME: self.global_config_name,
        }

        slurm_output_directory = self.data_io.get_data_path(SlurmOutputStd,data_args = data_args)
        self.data_io.create_directory(slurm_output_directory)
        script_buffer.write(f'#SBATCH -o {slurm_output_directory}\n')

        slurm_error_directory = self.data_io.get_data_path(SlurmOutputErr,data_args = data_args)
        script_buffer.write(f'#SBATCH -e {slurm_error_directory}\n')
        script_buffer.write(f'eval "$(conda shell.bash hook)"\n')
        script_buffer.write(f'conda activate {self.data_io.get_base_conda_name()}\n')

        if self.has_grid_job_params:
            grid_array_lengths = [len(array) for array in self.all_joined_grid_job_params.values()]
            for i in range(len(grid_array_lengths)):
                
                modulus= grid_array_lengths[i]
                divisor = 1
                for j in range(i + 1,len(grid_array_lengths)):
                    divisor *= grid_array_lengths[j]

                script_buffer.write( f'i_{i}=$(( ($SLURM_ARRAY_TASK_ID / {divisor}) % {modulus} ))\n')

                if self.grid_job_readable_labels:
                    #Add in the logic to get the readable labels for the grid job parameters
                    for j in range(modulus):
                        if j == 0:
                            script_buffer.write('if ')
                        else:
                            script_buffer.write('elif ')
                        
                        script_buffer.write(f'[ $i_{i} -eq {j} ]\n')
                        script_buffer.write(f'then\n')
                        readable_label = Config._get_parameter_for_config_name(
                            self.all_joined_grid_job_params,
                            i, j)
                        script_buffer.write(f'\ti_{i}=\"{readable_label}\"\n')
                    script_buffer.write('fi\n')
            config_file = self.data_io.get_data_path(Config, data_args={
                Config.CONFIG_NAME: self.global_config_name,
                Config.GRID_IDX: [f'${{i_{i}}}' for i in range(len(grid_array_lengths))],
            })
            self._write_command_into_script(script_buffer, config_file)

            total_array_length = 1
            for array_length in grid_array_lengths:
                total_array_length *= array_length
        else:
            config_file = self.data_io.get_data_path(Config, data_args={
                Config.CONFIG_NAME: self.global_config_name,
            })
            self._write_command_into_script(script_buffer, config_file)
            total_array_length = 1


        log_dir_name = os.path.dirname(slurm_output_directory)
        return self._save_script_and_return_path(script_buffer,GridSlurmScript) ,total_array_length, log_dir_name

    def config_info_iterator(self):
        data_args = {
            Config.CONFIG_NAME: self.global_config_name,
        }
        if self.grid_job_readable_labels:
            data_args[Config.GRID_IDX_LABELS] = self.all_joined_grid_job_params

        for array_combo, grid_indices in zip(product(*self.all_joined_grid_job_params.values()), product(*[range(len(array)) for array in self.all_joined_grid_job_params.values()])):
            grid_args = dict(zip(self.all_joined_grid_job_params.keys(), array_combo))
            data_args[Config.GRID_IDX]=grid_indices
            yield data_args, grid_args

    def build_configs(self):
        if not self.has_grid_job_params:
            return
        for config_data_args, grid_args in self.config_info_iterator():
            override_config = OmegaConf.create({
                'defaults': [self.global_config_name],
            })
            for config_keys, config_values in grid_args.items():
                for config_key, config_value in zip(list(config_keys), config_values):
                    if config_key == 'defaults':
                        override_config['defaults'] = [config_value] + override_config['defaults']
                    else:
                        #TODO: account for the case where a list config is passed, not just a dict. This can occur when there are many models modules. 
                        key_split = config_key.split('.')
                        nested_config_condition = {}
                        current = nested_config_condition
                        for key in key_split[:-1]:  
                            current = current.setdefault(key, {}) 
                        current[key_split[-1]] = config_value
                        override_config = OmegaConf.merge(override_config, OmegaConf.create(nested_config_condition))

            config_data_args[Config.DATA]= override_config
            self.data_io.save_data(Config, data_args = config_data_args)
            config_file_name = self.data_io.get_data_path(Config, data_args = config_data_args)
            self.config_list.append(config_file_name)        

    def build_and_execute_sequential_shell_script(self):
        shell_script_file_name = self.build_shell_script()
        command = f'./{shell_script_file_name}'
        self.data_io.run_system_command(command)

    def build_shell_script(self):
        script_buffer = io.StringIO()
        script_buffer.write(f'#!/bin/bash\n')
        for config_file in self.config_list:
            self._write_command_into_script(script_buffer, config_file)
        return self._save_script_and_return_path(script_buffer,GridShellScript)
    

    def _write_command_into_script(self,script_buffer, config_file):

        for action_code in self.grid_actions:
            if (action_code == ReservedArgKeywords.INFER.value) or \
                (action_code == ReservedArgKeywords.INFER_SHORT.value) or \
                (action_code == ReservedArgKeywords.TRAIN.value) or \
                (action_code == ReservedArgKeywords.TRAIN_SHORT.value):


                #/BRAIN/mmbt/static00/envs/mmbt/bin/accelerate
                base_conda_name = self.data_io.get_base_conda_name()
                script_buffer.write(f'{base_conda_name}/bin/accelerate launch')
                if self.grid_job_style == GridJobExecutionStyle.SLURM.value:
                    gpu = self.slurm_config.get('gpu', 0)
                    if gpu > 1:
                        script_buffer.write(' --multi_gpu')
                        print('WARNING: Using multi_gpu mode with SLURM, you will likely encounter port in use errors.')
                        print('Support for this is not implemented yet.')
                        print('A partial solution is to use the  --main_process_port to specify a port that is not in use.')
                        print('However, dev is still on going')
            else:
                script_buffer.write('python')

            script_buffer.write(f' main.py --{action_code} -c {config_file}')
            if self.has_grid_debug:
                script_buffer.write(f' --{ReservedArgKeywords.DEBUG.value}')
            script_buffer.write('\n')
            script_buffer.write(f'echo ""\n')
    
    def _save_script_and_return_path(self,script_buffer,script_class):
        script = script_buffer.getvalue()
        script_buffer.close()
        data_args = {
            script_class.CONFIG_NAME: self.global_config_name,
            script_class.DATA: script}
        
        self.data_io.save_data(script_class, data_args = data_args)
        script_file_name = self.data_io.get_data_path(script_class, data_args = data_args)
        return script_file_name
    
    def get_count_of_grid_jobs(self):
        if not self.has_grid_job_params:
            return 1
        return len(self.config_list)


class GridJobViewer(GridJobBuilder):
    def __init__(self,args):
        super().__init__(args)

    def __call__(self):
        self.print_slurm_outputs()
        self.print_slurm_errors()


    def _get_latest_timestamp(self):
        if self.grid_job_style != GridJobExecutionStyle.SLURM.value:
            raise Exception(f'Grid job viewer can only be used with SLURM style grid jobs, but got {self.grid_job_style}')
        latest_directory = None
        latest_timestamp = None
        for slurm_output_directory, data_args in self.data_io.get_file_iterator(AbstractSlurmOutput,
            data_args = {
                AbstractSlurmOutput.CONFIG_NAME: self.global_config_name,
                AbstractSlurmOutput.TIMESTAMP: FilePathEntries.OPEN_ENTRY},
            can_return_data_args= True,
            ):
            latest_directory = slurm_output_directory
            latest_timestamp = data_args[AbstractSlurmOutput.TIMESTAMP]

        if latest_directory is None:
            raise Exception(f'No SLURM output directory found for config {self.global_config_name}. Please run the grid job first.')

        return latest_timestamp
        
    def print_slurm_outputs(self):
        latest_timestamp = self._get_latest_timestamp()
        print('SLURM outputs:')
        for slurm_output in self.data_io.get_file_iterator(SlurmOutputStd, data_args = {
            SlurmOutputStd.CONFIG_NAME: self.global_config_name,
            SlurmOutputStd.TIMESTAMP: latest_timestamp,
            SlurmOutputStd.SLURM_FORMAT_FILENAME: FilePathEntries.OPEN_ENTRY}):
            print(slurm_output)

    def print_slurm_errors(self):
        print('SLURM errors:')
        latest_timestamp = self._get_latest_timestamp()
        for slurm_error in self.data_io.get_file_iterator(SlurmOutputErr, data_args = {
            SlurmOutputErr.CONFIG_NAME: self.global_config_name,
            SlurmOutputErr.TIMESTAMP: latest_timestamp,
            SlurmOutputErr.SLURM_FORMAT_FILENAME: FilePathEntries.OPEN_ENTRY}):
            yield slurm_error



class SequentialGridJobExecutor(DataIOObject):
    def __init__(self,args):
        super().__init__(args)
        self.check_every = 600 # Check every ten minutes for job completion 10 minutes
        self.all_configs = args.sequential_grid_job
        self.save_args = args

    def __call__(self):
        for config_name in self.all_configs:
            self.save_args.config_name = config_name
            grid_job_builder = GridJobBuilder(self.save_args)
            if grid_job_builder.grid_job_style != GridJobExecutionStyle.SLURM.value:
                raise Exception(f'Sequential grid job executor can only be used with SLURM style grid jobs (for now...). ')
            
            grid_job_builder()
            total_config_count = grid_job_builder.get_count_of_grid_jobs()
            slurm_job_indices = [-1] if total_config_count == 1 else list(range(total_config_count))
            slurm_job_id = grid_job_builder.slurm_job_id

            latest_timestamp = GridJobViewer(self.save_args)._get_latest_timestamp()


            while True:
                time.sleep(self.check_every)
                is_slurm_job_complete = grid_job_builder.is_slurm_job_complete()
                #Check if slurm job is off
                if is_slurm_job_complete:
                    for job_index in slurm_job_indices:
                        slurm_file_args = {
                            AbstractSlurmOutput.CONFIG_NAME: config_name,
                            AbstractSlurmOutput.SLURM_JOB_ID: slurm_job_id,
                            AbstractSlurmOutput.SLURM_ARRAY_TASK_ID: job_index,
                            AbstractSlurmOutput.TIMESTAMP: latest_timestamp,
                        }

                        if not self.data_io.is_file_present(SlurmOutputStd, data_args = slurm_file_args):
                            missing_filename = self.data_io.get_data_path(SlurmOutputStd, data_args = slurm_file_args)
                            raise Exception(f'SLURM job {job_index} is complete, but the output file {missing_filename} is missing. This may indicate that the job did not run or failed to produce output.')

                        
                        slurm_file_args[SlurmOutputStd.LAST_LINES] = 1
                        last_line = self.data_io.load_data(SlurmOutputStd, data_args = slurm_file_args)
                        if last_line != MainLoopMessages.COMPLETE_PROGRAM.value:
                            slurm_file_args[SlurmOutputErr.LAST_LINES] = 10
                            if self.data_io.is_file_present(SlurmOutputErr, data_args = slurm_file_args):
                                last_error_lines = self.data_io.load_data(SlurmOutputErr, data_args = slurm_file_args)
                                
                                print('last line was:', last_line)
                                if job_index == -1:
                                    print(f'last error lines for {slurm_job_id}:')
                                else:
                                    print(f'last error lines for {slurm_job_id} job index {job_index}:')
                                for line in last_error_lines:
                                    print(line)
                            else:
                                missing_error_filename = self.data_io.get_data_path(SlurmOutputErr, data_args = slurm_file_args)
                                print(f'No error log {missing_error_filename} produced.')                            
                            raise Exception(f'SLURM job for config {config_name} did not complete successfully. See the above lines for details')
                    break



import argparse
import random
from file_golem import FileGolem

from ml_golem.model_action import ModelAction
from ml_golem.overleaf.overleaf_transfer import OverleafTransfer
from ml_golem.grid_search.tensorboard_logic import TensorBoardLogic
from ml_golem.grid_search.config_copier import ConfigCopier
from ml_golem.grid_search.grid_search_builder import GridJobBuilder, GridJobViewer,SequentialGridJobExecutor
from ml_golem.main_loop_logic.reserved_keywords import ReservedArgKeywords


def add_flag_argument(parser,action_list,object_type,flag_name,flag_abbreviation=None,help_text=None):
    if help_text is None:
        help_text = f'Use this flag to enable {flag_name}'

    if flag_abbreviation is None:
        parser.add_argument(f'--{flag_name}',action='store_true',help=help_text)
    else:
        parser.add_argument(f'--{flag_name}',f'-{flag_abbreviation}',action='store_true',help=help_text)

    action = lambda args: (object_type(args)() or True) if getattr(args, flag_name) else False
    action_list.append(action)
    return parser, action_list


def get_args_and_actions(app_title,system_config_path,_add_additional_args):
    action_list = []
    parser = argparse.ArgumentParser(description=app_title)

    parser.add_argument('--data_io',default = None, help='Data IO object to use')
    parser.add_argument('-d',f'--{ReservedArgKeywords.DEBUG.value}',action='store_true',help='Print debug statements')
    parser.add_argument('-st','--system_transfer',type=str, default=None, help='System that the data_io object is transfering into' )

    parser.add_argument('-s','--seed', type=int,default=random.randint(0, 2**32 - 1), help='Seed for random number')
    parser.add_argument('-c','--config_name', type=str, default=None, help='Path to relevant omega conf file')

    parser.add_argument('-cc','--copy_config', type=str, default=None, help='Copy a config file')
    action_list.append(lambda args: (ConfigCopier(args)() or True) if args.copy_config else False)


    parser.add_argument(f'-{ReservedArgKeywords.TRAIN_SHORT.value}', f'--{ReservedArgKeywords.TRAIN.value}', 
        action= 'store_true', help='Train the model')
    action_list.append(lambda args: (ModelAction(args).train(args) or True) if args.train else False)

    parser.add_argument(f'-{ReservedArgKeywords.INFER_SHORT.value}', f'--{ReservedArgKeywords.INFER.value}', action= 'store_true', help='Have the model make inferences')
    action_list.append(lambda args: (ModelAction(args).inference(args) or True) if args.infer else False)

    parser.add_argument('-w', '--wipe', action= 'store_true', help='Wipe the model and training logs')
    action_list.append(lambda args: (ModelAction(args).wipe() or True) if args.wipe else False)

    parser.add_argument('-g', '--grid_job', action= 'store_true', help='Run a grid job')
    action_list.append(lambda args: (GridJobBuilder(args)() or True) if args.grid_job else False)

    parser.add_argument('-vg','--view_grid_job', action='store_true', help='View the grid job')
    action_list.append(lambda args: (GridJobViewer(args)() or True) if args.view_grid_job else False)

    parser.add_argument(
        f'-{ReservedArgKeywords.SEQUENTIAL_GRID_JOB_SHORT.value}',
        f'--{ReservedArgKeywords.SEQUENTIAL_GRID_JOB.value}',
        nargs='+', 
        type=str,
        default=[], help='Run a sequence of grid jobs. Provide the paths to the config files as arguments.')
    action_list.append(lambda args: (SequentialGridJobExecutor(args)() or True) if len(args.sequential_grid_job) > 0 else False)

    parser.add_argument('-co', '--call_object', type=str, default=None, help='construct an object with args and call __call__')
    action_list.append(lambda args: (args.data_io.fetch_class(args.call_object)(args)() or True) if args.call_object else False)

    parser.add_argument('-co_args', '--call_object_args', type=str, default=None, help='Arguments to pass to the call_object')

    parser.add_argument('-fi', '--find_import', type=str, default=None, help='Find the import path for a given class')
    action_list.append(lambda args: (args.data_io.find_import(args.find_import) or True) if args.find_import else False)

    parser.add_argument('-ot', '--overleaf_transfer', action='store_true', help='Update git-linked Overleaf to contain the latest graphics')
    action_list.append(lambda args: (OverleafTransfer(args)() or True) if args.overleaf_transfer else False)

    parser.add_argument('-tb', '--tensorboard', action='store_true', help='run the tensorboard for a given config')
    action_list.append(lambda args: (TensorBoardLogic(args)() or True) if args.tensorboard else False)


    if _add_additional_args is not None:
        parser, action_list = _add_additional_args(parser,action_list)

    args = parser.parse_args()
    #if args.call_object_args is not None:
    co_args_string = [] if args.call_object_args is None else args.call_object_args.split(',')
    args.call_object_args = {}
    for arg in co_args_string:
        if '=' in arg:
            key, value = arg.split('=', 1)
            args.call_object_args[key.strip()] = value.strip()
        else:
            args.call_object_args[arg.strip()] = None

    if 'help' in args:
        parser.print_help()

    if 'local_config_name' in args:
        raise Exception(f'local_config_name is a reserved keyword in args, it is set as {args.local_config_name} but should be empty')
    
    if 'origin_config_name' in args:
        raise Exception(f'origin_config_name is a reserved keyword in args, it is set as {args.origin_config_name} but should be empty')

    if 'object_ancestry' in args:
        raise Exception(f'object_ancestry is a reserved keyword in args, it is set as {args.object_ancestry} but should be empty')

    if (ReservedArgKeywords.INFER.value) in (args and ReservedArgKeywords.TRAIN.value) in args:
        raise Exception(f'You cannot use both {ReservedArgKeywords.INFER.value} and {ReservedArgKeywords.TRAIN.value} at the same time. Please choose one.')

    args.data_io =FileGolem(
        system_config_path=system_config_path,
        is_debug=args.debug,
        system_transfer = args.system_transfer
    )

    return args, action_list



from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass

class LossBase(ConfigBasedClass):
    MAIN_LOSS = 'main_loss'
    def __init__(self,args,subconfig_keys):
        super().__init__(args,subconfig_keys)
        self.batch_loss_dict = {
            self.MAIN_LOSS: []
        }


    def __call__(self, model_input, model_output):
        raise Exception('Not Implemented')
    

    def store_batch_loss(self,loss_results):
        for loss_key in loss_results.keys():
            if loss_key not in self.batch_loss_dict:
                self.batch_loss_dict[loss_key] = []
            self.batch_loss_dict[loss_key].append(loss_results[loss_key].item())

        #self.batch_main_losses.append(loss_results[self.MAIN_LOSS])

    def log_epoch_loss(self, writer, epoch, loss_prefix=''):
        """
        Log the total loss for the epoch to TensorBoard.
        """
        for key in self.batch_loss_dict.keys():
            if len(self.batch_loss_dict[key]) == 0:
                continue
            mean_batch_loss = sum(self.batch_loss_dict[key]) / len(self.batch_loss_dict[key])
            writer.add_scalar(f'{loss_prefix} Loss/{key}', mean_batch_loss, epoch)
            self.batch_loss_dict[key] = [] 


from file_golem import FileDatatypes,AbstractDatatype,FilePathEntries
from file_golem import Config, SpecialDataArgs

class ModelCheckpoint(AbstractDatatype):
    FILE_DATATYPE = FileDatatypes.TORCH_CHECKPOINT
    EPOCH = 'epoch'
    MODULE = 'module'
    RELATIVE_PATH_TUPLE = AbstractDatatype.RELATIVE_PATH_TUPLE + (
        'model_checkpoints',
        FilePathEntries.CONFIG_ENTRY,
        {FilePathEntries.CUSTOM_LOGIC: '_include_module_entry'},
        {FilePathEntries.DATA_ARG_ENTRY: EPOCH})
    
    @staticmethod
    def _include_module_entry(data_args):
        if ModelCheckpoint.MODULE not in data_args:
            return None
        module = data_args[ModelCheckpoint.MODULE]
        if module == FilePathEntries.OPEN_ENTRY:
            return '*'
        return module

    @staticmethod
    def _retrieve_data_args(relative_path_tuple, data_args):
        args_to_return = {
            ModelCheckpoint.EPOCH: relative_path_tuple[-1],
        }
        return args_to_return


class TrainingLog(AbstractDatatype):
    FILE_DATATYPE = FileDatatypes.EMPTY
    RELATIVE_PATH_TUPLE = AbstractDatatype.RELATIVE_PATH_TUPLE + (
        'training_logs',
        FilePathEntries.CONFIG_ENTRY)



class EvaluationResults(AbstractDatatype):
    FILE_DATATYPE = FileDatatypes.JSON
    RELATIVE_PATH_TUPLE = AbstractDatatype.RELATIVE_PATH_TUPLE + (
        'evaluation_log',)


class GridShellScript(Config):
    FILE_DATATYPE = FileDatatypes.SHELL
    IS_EXECUTABLE = True  # Shell scripts are executable
    RELATIVE_PATH_TUPLE = Config.RELATIVE_PATH_TUPLE + ('_sequential_job',)


class GridSlurmScript(Config):
    FILE_DATATYPE = FileDatatypes.SLURM_SCRIPT
    IS_EXECUTABLE = True  # Slurm scripts are executable
    RELATIVE_PATH_TUPLE = Config.RELATIVE_PATH_TUPLE + ('_slurm_job',)


### SLURM OUTPUTS
class AbstractSlurmOutput(AbstractDatatype):
    SLURM_FORMAT_FILENAME = 'slurm_format_filename'
    FILE_DATATYPE = FileDatatypes.EMPTY
    LAST_LINES = 'last_lines'
    SLURM_JOB_ID = 'slurm_job_id'
    SLURM_ARRAY_TASK_ID = 'slurm_array_task_id'
    RELATIVE_PATH_TUPLE = AbstractDatatype.RELATIVE_PATH_TUPLE + (
        'misc',
        'slurm',
        FilePathEntries.TIMESTAMP_CONFIG_ENTRY,)
        #'%j')
        # 
    @staticmethod
    def _retrieve_data_args(relative_path_tuple, data_args):
        config_timestamp_entry = relative_path_tuple[-1]
        timestamp = config_timestamp_entry[-data_args[SpecialDataArgs.DATA_IO]._get_timestamp_format_size():]
        data_args[AbstractSlurmOutput.TIMESTAMP] = timestamp
        return data_args
    
    @staticmethod
    def _slurm_format_or_open_entry(data_args):

        if AbstractSlurmOutput.SLURM_JOB_ID in data_args:
            job_id = data_args[AbstractSlurmOutput.SLURM_JOB_ID]
            if job_id == FilePathEntries.OPEN_ENTRY:
                job_prefix = '*'
            else:
                job_prefix = f'{job_id}'
            if AbstractSlurmOutput.SLURM_ARRAY_TASK_ID in data_args:
                task_id = data_args[AbstractSlurmOutput.SLURM_ARRAY_TASK_ID]
                if task_id == FilePathEntries.OPEN_ENTRY:
                    job_index_suffix = '_*'
                elif task_id == -1:
                    job_index_suffix = ''
                else:
                    job_index_suffix = f'_{task_id}'
            return f'{job_prefix}{job_index_suffix}'

            
        if AbstractSlurmOutput.SLURM_FORMAT_FILENAME in data_args:
            file_name = data_args[AbstractSlurmOutput.SLURM_FORMAT_FILENAME]
            if file_name != FilePathEntries.OPEN_ENTRY:
                raise Exception('Can only use slurm format filename with open entry')
            return '*'
        else:
            return '%A_%a'
    
class SlurmOutputStd(AbstractSlurmOutput):
    FILE_DATATYPE = FileDatatypes.SLURM_OUTPUT_STD
    RELATIVE_PATH_TUPLE = AbstractSlurmOutput.RELATIVE_PATH_TUPLE + (
         {FilePathEntries.CUSTOM_LOGIC: '_slurm_format_or_open_entry'},)


class SlurmOutputErr(AbstractSlurmOutput):
    FILE_DATATYPE = FileDatatypes.SLURM_OUTPUT_ERR
    RELATIVE_PATH_TUPLE = AbstractSlurmOutput.RELATIVE_PATH_TUPLE + (
         {FilePathEntries.CUSTOM_LOGIC: '_slurm_format_or_open_entry'},)



import torch
from tqdm import tqdm
from torch.utils.tensorboard import SummaryWriter
from ml_golem.base_classes.model_io_base import ModelIOBase,ModelConfigKeywords
from ml_golem.datatypes import TrainingLog
from ml_golem.base_classes._timer import Timer

class ModelTrainer(ModelIOBase):
    def __init__(self,args,subconfig_keys):
        super().__init__(args,subconfig_keys)


        self.loss = self._initialize_loss(args,subconfig_keys)
        self.optimizer = self._initialize_optimizer()
        
        self.save_every = self.config[ModelConfigKeywords.SAVE_EVERY.value]

        #TODO: Improve the logic to have validation inherent from training, unless overrides are specified.
        self.validate_every = self.config.get(ModelConfigKeywords.VALIDATE_EVERY.value, -1)
        if self.validate_every > 0:
            self.validation_dataloader = self._initialize_dataloader(args,subconfig_keys+[ModelConfigKeywords.VALIDATION.value])
            self.validation_loss = self._initialize_loss(args,subconfig_keys+[ModelConfigKeywords.VALIDATION.value])

        self._prepare_accelerator()
        

        #TODO: DIG DEEPER INTO using .add_hparams to store relevant hyperparameters on grid searches
        self.writer = SummaryWriter(self.data_io.get_data_path(TrainingLog,data_args ={
            TrainingLog.CONFIG_NAME: self.global_config_name,
        }))
        self.epochs = self.config[ModelConfigKeywords.EPOCHS.value]
        self.can_display_epoch_progress = self.config.get(ModelConfigKeywords.CAN_DISPLAY_EPOCH_PROGRESS.value, True)
        self.can_time_batch = self.config.get(ModelConfigKeywords.CAN_TIME_BATCH.value, False)
        if self.can_time_batch:
            self.timer = Timer()
            self._batch_size = self.data_io.fetch_subconfig(self.global_config_name,
                subconfig_keys=subconfig_keys+[ModelConfigKeywords.DATALOADER.value]).get(ModelConfigKeywords.BATCH_SIZE.value, 1)


    def __call__(self):
        self._call_core_function_with_autocast(self._train_core)


    def _train_core(self):
        #TODO: consider adding support for torch.cuda.amp.autocast for mixed precision training.
        self.switch_model_to_train()  # Switch back to training mode
        epoch_iterator = range(self.resume_epoch, self.epochs)
        if self.can_display_epoch_progress:
            epoch_iterator = tqdm(epoch_iterator, desc="Epochs")
        for epoch in epoch_iterator:
            self._check_epoch_to_save_or_validate(epoch)
            for input_batch in self.dataloader:
                if self.can_time_batch:
                    self.timer.start()
                self.optimizer.zero_grad()
                output_batch = self.make_forward_pass(input_batch)
                loss_results = self.loss(input_batch,output_batch)
                self.accelerator.backward(loss_results[self.loss.MAIN_LOSS])
                self.optimizer.step()
                self.loss.store_batch_loss(loss_results)
                if self.can_time_batch:
                    elapsed_time = self.timer.stop()
                    predicted_epoch_time = elapsed_time * len(self.dataloader) / (self._batch_size)
                    print(f'Batch time: {self.timer.format_time(elapsed_time)}. Predicted epoch time: {self.timer.format_time(predicted_epoch_time)}')
            self.loss.log_epoch_loss(self.writer,epoch,loss_prefix='train_')
        self._check_epoch_to_save_or_validate(self.epochs)
        self.writer.close() #Might be necessary to close the writer to ensure all data is flushed to disk.

    def _can_shuffle(self):
        return True
    
    def _check_epoch_to_save_or_validate(self,epoch):
        if (epoch % self.save_every == 0) or (epoch == self.epochs):
            self.save_model_on_epoch(epoch)

        if self.validate_every > 0:
            if (epoch % self.validate_every == 0) or (epoch == self.epochs):
                self.validate_on_epoch(epoch)

    def validate_on_epoch(self, epoch):
        print(f'Validating on epoch {epoch}')
        #self.model.eval()  # Set model to evaluation mode
        self.switch_model_to_eval()  # Switch to evaluation mode
        with torch.no_grad():  # Disable gradient computation
            for input_batch in self.validation_dataloader:
                output_batch = self.make_forward_pass(input_batch)
                loss_results = self.validation_loss(input_batch, output_batch)
                self.validation_loss.store_batch_loss(loss_results)
            self.validation_loss.log_epoch_loss(self.writer,epoch,loss_prefix='validate_')

        #self.model.train()  # Set model back to training mode
        self.switch_model_to_train()  # Switch back to training mode
        print(f'Validation of epoch {epoch} complete')


    def save_model_on_epoch(self, epoch):
        if self.accelerator.is_main_process:
            self.save_model_checkpoint(epoch)
            # num_gpus = self.accelerator.state.num_processes
            # if num_gpus > 1:
            #     original_model = self.model.module
            # else:
            #     original_model = self.model
            # self.save_model_checkpoint(original_model,epoch)
        self.accelerator.wait_for_everyone()


from enum import Enum
class ModelConfigKeywords(Enum):
    MODEL_CLASS = 'model_class'
    TRAINING = 'training'
    INFERENCE = 'inference'
    ARCHITECTURE = 'architecture'
    CONFIG = 'config'
    WEIGHT_DECAY = 'weight_decay'
    ORIGIN_CONFIG = 'origin_config'
    ORIGIN_MODULE = 'origin_module'


    RESUME_EPOCH = 'resume_epoch'

    #TRAINING KEYWORDS
    EPOCHS = 'epochs'
    LEARNING_RATE = 'learning_rate'
    CAN_TIME_BATCH = 'can_time_batch'
    SAVE_EVERY = 'save_every'
    VALIDATE_EVERY = 'validate_every'
    VALIDATION = 'validation'
    CAN_DISPLAY_EPOCH_PROGRESS = 'can_display_epoch_progress'

    DATASET = 'dataset'
    CAN_SHUFFLE = 'can_shuffle'

    MODEL_OUTPUT = 'model_output'
    GROUND_TRUTH = 'ground_truth'
    DATATYPE = 'datatype'
    DATA_SOURCES = 'data_sources'

    NUM_WORKERS = 'num_workers'
    BATCH_SIZE = 'batch_size'
    DATALOADER = 'dataloader'
    LOSS = 'loss'

    IS_PRELOADED = 'is_preloaded'
    HAS_INDEX = 'has_index'

    AUTOCASTING = 'autocasting'
    CAN_AUTOCAST = 'can_autocast'
    AUTOCAST_DTYPE = 'autocast_dtype'  # e.g., 'float16', 'bfloat16'
    AUTOCAST_CACHE_ENABLED = 'autocast_cache_enabled'


    GRID_JOB = 'grid_job'
    GRID_JOB_PARAMS = 'grid_job_params'
    GRID_ACTIONS = 'grid_actions'
    GRID_JOB_STYLE = 'grid_job_style'
    GRID_SLURM_CONFIG = 'grid_slurm_config'
    GRID_DEBUG = 'grid_debug'
    GRID_JOB_JOINS = 'grid_job_joins'
    GRID_JOB_READABLE_LABELS = 'grid_job_readable_labels'


from .main_loop_logic.main_loop import main_loop

from .model_loading_logic.config_based_class import ConfigBasedClass
from .model_loading_logic.model_config_keywords import ModelConfigKeywords


from .base_classes.data_io_object import DataIOObject
from .base_classes.dataset_base import DatasetBase
from .base_classes.model_base import ModelBase
from .base_classes.loss_base import LossBase
from .model_inference import ModelInference
from .model_trainer import ModelTrainer



import torch
from torch.utils.data import Dataset
from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass
from ml_golem.model_loading_logic.model_config_keywords import ModelConfigKeywords
from file_golem import FileDatatypes


class DatasetBase(ConfigBasedClass,Dataset):
    DEFAULT_DATASOURCE = 'default_datasource'

    def __init__(self,args,subconfig_keys):
        super().__init__(args,subconfig_keys)

        self.has_index = self.config.get(ModelConfigKeywords.HAS_INDEX.value,False)
        self.is_preloaded = self.config.get(ModelConfigKeywords.IS_PRELOADED.value, False)


        datasources= self.config.get(ModelConfigKeywords.DATA_SOURCES.value, {})
        self.datasources = {}
        if len(datasources) == 0:
            self.datasources[self.DEFAULT_DATASOURCE] = self.extract_dataset_config_and_datatype(self.config,self.DEFAULT_DATASOURCE)
        else:
            for key in datasources.keys():
                datasource_config = datasources[key]
                self.datasources[key] = self.extract_dataset_config_and_datatype(datasource_config,key)
        self.db = {}


    def _default_config(self):
        defaults = {
            self.DEFAULT_DATASOURCE: {
                ModelConfigKeywords.CONFIG.value: None,
                ModelConfigKeywords.DATATYPE.value: None
            }
        }
        return defaults


    def extract_dataset_config_and_datatype(self,config,datasource_key):
        all_dataset_config_and_datatype = self._default_config()

        dataset_config = config.get(ModelConfigKeywords.CONFIG.value, None)
        if dataset_config is None:
            if datasource_key in all_dataset_config_and_datatype:
                dataset_config = all_dataset_config_and_datatype[datasource_key][ModelConfigKeywords.CONFIG.value]
            
        datatype = self.data_io.fetch_class_from_config(
            config_or_config_name = config,
            subconfig_keys=[ModelConfigKeywords.DATATYPE.value],
            is_required=False)
        if datatype is None:
            if datasource_key in all_dataset_config_and_datatype:
                datatype = all_dataset_config_and_datatype[datasource_key][ModelConfigKeywords.DATATYPE.value]

        data_dict = {
            ModelConfigKeywords.CONFIG.value: dataset_config,
            ModelConfigKeywords.DATATYPE.value: datatype
        }
        return data_dict



    def _load_item(self,idx):
        dataset_config = self._get_dataset_config()
        datatype = self._get_datatype()
        data_item = self.data_io.load_data(datatype,data_args={
            datatype.IDX: idx,
            datatype.CONFIG_NAME: dataset_config})
        
        if self.has_index:
            data_item[datatype.IDX] = torch.tensor(idx)
        
        return data_item
    
    def __len__(self):
        dataset_config = self._get_dataset_config()
        datatype = self._get_datatype()
        return self.data_io.get_datatype_length(datatype,data_args= {
            datatype.CONFIG_NAME: dataset_config
        })


    def _get_custom_collate_fn(self):
        if hasattr(self,'custom_collate_fn'):
            return self.custom_collate_fn
        
        datatype = self._get_datatype()
        if datatype is None or (datatype.FILE_DATATYPE != FileDatatypes.TORCH):
            return lambda x: x    
        return None


    def __getitem__(self, idx):
        if self.is_preloaded:
            if len(self.db)==0:
                for i in range(len(self)):
                    x = self._load_item(i)
                    self.db[i] = x

            return self.db[idx]
        else:
            return self._load_item(idx)
        

    def _get_dataset_config(self,datasource=None):
        if len(self.datasources) == 1:
            return self.datasources[self.DEFAULT_DATASOURCE][ModelConfigKeywords.CONFIG.value]
        if datasource is not None:
            return self.datasources[datasource][ModelConfigKeywords.CONFIG.value]
        raise Exception('Multiple datasources found, please specify the datasource')
    

    def _get_datatype(self,datasource=None):
        if len(self.datasources) == 1:
            return self.datasources[self.DEFAULT_DATASOURCE][ModelConfigKeywords.DATATYPE.value]
        if datasource is not None:
            return self.datasources[datasource][ModelConfigKeywords.DATATYPE.value]
        return None
        #raise Exception('Multiple datasources found, please specify the datasource')



from enum import Enum

class ReservedArgKeywords(Enum):
    DATA_IO = 'data_io'
    DEBUG = 'debug'
    DEBUG_SHORT = 'd'
    SYSTEM_TRANSFER = 'system_transfer'
    SYSTEM_TRANSFER_SHORT = 'st'
    SEED = 'seed'
    CONFIG_NAME = 'config_name'
    CONFIG_NAME_SHORT = 'c'
    COPY_CONFIG = 'copy_config'
    COPY_CONFIG_SHORT = 'cc'
    TRAIN = 'train'
    TRAIN_SHORT = 't'
    INFER = 'infer'
    INFER_SHORT = 'i'
    WIPE = 'wipe'
    WIPE_SHORT = 'w'
    GRID_JOB = 'grid_job'
    GRID_JOB_SHORT = 'g'
    VIEW_GRID_JOB = 'view_grid_job'
    VIEW_GRID_JOB_SHORT = 'vg'
    SEQUENTIAL_GRID_JOB = 'sequential_grid_job'
    SEQUENTIAL_GRID_JOB_SHORT = 'sg'
    CALL_OBJECT = 'call_object'
    CALL_OBJECT_SHORT = 'co'
    CALL_OBJECT_ARGS = 'call_object_args'
    CALL_OBJECT_ARGS_SHORT = 'co_args'
    FIND_IMPORT = 'find_import'
    FIND_IMPORT_SHORT = 'fi'
    OVERLEAF_TRANSFER = 'overleaf_transfer'
    OVERLEAF_TRANSFER_SHORT = 'ot'
    TENSORBOARD = 'tensorboard'
    TENSORBOARD_SHORT = 'tb'


from setuptools import setup, find_packages

setup(
    name='ml_golem',
    version='0.1.6',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'file_golem',
        'torch',
        'accelerate',
        'tensorboard',
        'pandas'
    ],
    # entry_points={
    #     'console_scripts': [
    #         'ml_lib_main = ml_lib.main:main',
    #     ],
    # },
    author='Cameron Braunstein',
    author_email='your.email@example.com',
    description='A description of your project',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/CameronBraunstein/ml_golem',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)


from ml_golem.base_classes.data_io_object import DataIOObject

from ml_golem.model_loading_logic.model_config_keywords import ModelConfigKeywords


class ConfigBasedClass(DataIOObject):
    def __init__(self,args,subconfig_keys=[]):
        super().__init__(args)
        self.subconfig_keys = subconfig_keys
        self.global_config_name = args.config_name
        if 'local_config_name' in args:
            self.local_config_name = args.local_config_name
        else:
            self.local_config_name = self.global_config_name

        if 'origin_config_name' in args:
            self.origin_config_name = args.origin_config_name

        if 'object_ancestry' in args:
            self.object_ancestry = args.object_ancestry

        self.config = self.data_io.fetch_subconfig(
            self.local_config_name,
            subconfig_keys= subconfig_keys)
        
        
    def _get_ancestor(self):
        """
        Returns the ancestor object if it exists, otherwise returns self.
        This is useful for accessing the config of the ancestor object.
        """
        return self.object_ancestry[-1]
    
    def _get_local_config(self):
        """
        Returns the local config of the current object.
        This is useful for accessing the config of the current object.
        """
        return self.config

    def instantiate_config_based_class(self,args,config_or_config_name,subconfig_keys,default_class=None,origin_config=None):
        if len(subconfig_keys) == 0:
            print('A non-zero length list subconfig_keys must be provided to instantiate a config based class.')
            print('Not doing so will likely cause an infinite loop.')
            print('In this situation, it is best practice to add levels to your config')
            print('You can access a the instantiator classes config by calling ._get_ancestor()._get_local_config() ')
            raise Exception('Empty subconfig_keys provided to instantiate_config_based_class. See warnings above.')
        
        config_class =self.data_io.fetch_class_from_config(
            config_or_config_name,
            subconfig_keys = subconfig_keys + [ModelConfigKeywords.MODEL_CLASS.value],
            is_required = False)
        config_class = default_class if config_class is None else config_class

        if config_class is None:
            raise Exception(f'No class found for keys {subconfig_keys + [ModelConfigKeywords.MODEL_CLASS.value]} in config {config_or_config_name}')  


        args.origin_config_name =  self.global_config_name if origin_config is None else origin_config
        args.local_config_name = config_or_config_name
        if hasattr(self,'object_ancestry'):
            args.object_ancestry = [*self.object_ancestry,self]
        else:
            args.object_ancestry = [self]


        instantiated_class = config_class(args, subconfig_keys)

        if 'local_config_name' in args:
            del args.local_config_name

        if 'origin_config_name' in args:
            del args.origin_config_name

        if 'object_ancestry' in args:
            del args.object_ancestry
        return instantiated_class



from ml_golem.model_loading_logic.config_based_class import ConfigBasedClass
from ml_golem.model_loading_logic.model_config_keywords import ModelConfigKeywords
from ml_golem.datatypes import ModelCheckpoint, TrainingLog
from file_golem import FilePathEntries
from ml_golem.model_inference import ModelInference
from ml_golem.model_trainer import ModelTrainer

class ModelAction(ConfigBasedClass):
    def train(self,args):
        print(f'Training: {self.global_config_name}')
        trainer = self.instantiate_config_based_class(
            args,
            self.global_config_name,
            subconfig_keys = [ModelConfigKeywords.TRAINING.value],
            default_class=ModelTrainer)
        trainer()

    def inference(self,args):
        print(f'Inference: {self.global_config_name}')
        inference = self.instantiate_config_based_class(
            args,
            self.global_config_name,
            subconfig_keys = [ModelConfigKeywords.INFERENCE.value],
            default_class=ModelInference)
        inference()

    def wipe(self):
        print(f'Wipe: {self.global_config_name}')
        print('WARNING: YOU PROBABLY DONT WANT TO CALL WIPE')
        self.data_io.delete_data(ModelCheckpoint,data_args = {
            ModelCheckpoint.CONFIG_NAME: self.global_config_name,
            ModelCheckpoint.EPOCH:FilePathEntries.OPEN_ENTRY
        })

        self.data_io.delete_data(ModelCheckpoint,data_args = {
            ModelCheckpoint.CONFIG_NAME: self.global_config_name,
            ModelCheckpoint.EPOCH:FilePathEntries.OPEN_ENTRY,
            ModelCheckpoint.MODULE: FilePathEntries.OPEN_ENTRY
        })

        self.data_io.delete_data(TrainingLog,data_args = {
            TrainingLog.CONFIG_NAME: self.global_config_name,
        })


import torch
import torch.nn as nn
from accelerate import Accelerator
from file_golem import FilePathEntries
from ml_golem.datatypes import ModelCheckpoint
from ml_golem.model_loading_logic.model_config_keywords import ModelConfigKeywords
from ml_golem.base_classes.dataloading_base import DataIterationBase
from ml_golem.base_classes.model_base import ModelBase
from omegaconf.listconfig import ListConfig


##Possible patterns:
# architecture:
#    model_class: ...

#architecture: 
#   origin_config: <config_name>
#   (opt) origin_module: <module_1, or module_2, ...>
#   resume_epoch: <n>

# architecture:
#    -module_1:
#      model_class: ...
#    -module_2:
#      model_class: ...
#  ....

# architecture:
#    -module_1:
#        origin_config: <config_name>
#        resume_epoch: <n>
#        (opt) origin_module: <module_1, or module_2, ...>
#    -module_2: <config_name>
#        origin_config: <config_name>
#        resume_epoch: <n>
#        (opt) origin_module: <module_1, or module_2, ...>
# ..




class ModelIOBase(DataIterationBase):
    def __init__(self, args,subconfig_keys):
        super().__init__(args,subconfig_keys)
        ###Autocasting paramters
        self.can_autocast = self.data_io.fetch_config_field(self.config, subconfig_keys=[ModelConfigKeywords.AUTOCASTING.value, ModelConfigKeywords.CAN_AUTOCAST.value], is_required=False, default_value=False)
        self.autocast_dtype = self.data_io.fetch_config_field(self.config, subconfig_keys=[ModelConfigKeywords.AUTOCASTING.value,ModelConfigKeywords.AUTOCAST_DTYPE.value], is_required=False, default_value=None)
        self.autocast_cache_enabled = self.data_io.fetch_config_field(self.config, subconfig_keys=[ModelConfigKeywords.AUTOCASTING.value,ModelConfigKeywords.AUTOCAST_CACHE_ENABLED.value], is_required=False, default_value=True)


        self.dataloader = self._initialize_dataloader(args,self.subconfig_keys)

        architecture_info = self.data_io.fetch_config_field(
            self.global_config_name,
            subconfig_keys =[ModelConfigKeywords.ARCHITECTURE.value],
            is_required=True)
        
        if isinstance(architecture_info, str):
            print(f'warning: architecture: {architecture_info} is a deprecated format. Please change to: ')
            print(f'architecture:\n    origin_config: {architecture_info}')
            print(f'architecture:\n    origin_module: <module_1, or module_2, ...>')
            raise Exception('Please change the architecture format to the new format. See the documentation for more details.')
        
    
        if isinstance(architecture_info, ListConfig):
            self.model_module_names = [list(module.keys())[0] for module in architecture_info]
            #self.module_module_configs = [list(module.values())[0] for module in architecture_info]
        else:
            self.model_module_names = [None]
            #self.module_module_configs = [self.global_config_name]

        self.models = []
        self.is_external_configs = []
        instantiate_configs = []
        instantiate_module_names = []

        for module_name in self.model_module_names:
            origin_config, origin_module_name, is_external_config = self._handle_recursive_module_references(
                self.global_config_name,
                module_name
            )
            self.is_external_configs.append(is_external_config)
            subconfig_keys = self._create_origin_subconfig_keys(origin_module_name)
            model_module = self.instantiate_config_based_class(
                args,
                origin_config, 
                subconfig_keys = subconfig_keys,
                origin_config = origin_config)

            instantiate_configs.append(origin_config)
            instantiate_module_names.append(origin_module_name)
            self.models.append(model_module)

        self.resume_epoch = int(self.data_io.fetch_config_field(
            self.config,
            subconfig_keys = [ModelConfigKeywords.RESUME_EPOCH.value],
            is_required=False,
            default_value=-1))
        
        if self.resume_epoch == -1: #Look for the latest checkpoint

            #Find the correct resume epoch for the global config.
            for model, instantiate_module_name, is_external_config in zip(self.models, instantiate_module_names, self.is_external_configs):
                if is_external_config:
                    continue
                if isinstance(model, nn.Module):
                    for _, file_args in self.data_io.get_file_iterator(
                        ModelCheckpoint,
                        data_args = {
                            ModelCheckpoint.CONFIG_NAME: self.global_config_name,
                            ModelCheckpoint.EPOCH: FilePathEntries.OPEN_ENTRY,
                            ModelCheckpoint.MODULE: instantiate_module_name,
                        },
                        can_return_data_args=True):
                        new_epoch = int(file_args[ModelCheckpoint.EPOCH])
                        if new_epoch > self.resume_epoch:
                            self.resume_epoch = new_epoch
                    break #We found the latest checkpoint for the global config, no need to check other modules

        for model,instantiate_config, instantiate_module_name, is_external_config, module_name \
            in zip(self.models, instantiate_configs, instantiate_module_names, self.is_external_configs, self.model_module_names):
            if not isinstance(model, nn.Module):
                self._print_skipping_loading_message(module_name, instantiate_config)
                continue

            if (self.resume_epoch == -1) and is_external_config: #In this case we insist on the user specifying the resume epoch. This could be made more flexible in the future.
                subconfig_keys = self._create_origin_subconfig_keys(module_name)
                subconfig_keys.append(ModelConfigKeywords.RESUME_EPOCH.value)
                module_resume_epoch = self.data_io.fetch_config_field(
                    self.global_config_name,
                    subconfig_keys = subconfig_keys,
                    is_required=True)
            else:
                module_resume_epoch = self.resume_epoch

            self._print_resume_epoch_message(module_name,instantiate_config, module_resume_epoch)

            if module_resume_epoch != -1:
                model_checkpoint = self.data_io.load_data(ModelCheckpoint, data_args = {
                    ModelCheckpoint.CONFIG_NAME: instantiate_config, #self.global_config_name,
                    ModelCheckpoint.MODULE: instantiate_module_name,
                    ModelCheckpoint.EPOCH: module_resume_epoch
                })
                model.load_state_dict(model_checkpoint)
            
            
            model._set_resume_epoch(max(0,self.resume_epoch))
        
        self.resume_epoch = max(0,self.resume_epoch) #Ensure the resume epoch is not negative
        print(f'Model run will begin at epoch {self.resume_epoch}')

    def _call_core_function_with_autocast(self,func):
        if self.can_autocast:
            if self.autocast_dtype is None:
                with torch.amp.autocast(self.accelerator.device.type, enabled=self.autocast_cache_enabled):
                    func()
            else:
                dtype = getattr(torch, self.autocast_dtype)
                with torch.amp.autocast(self.accelerator.device.type, enabled=self.autocast_cache_enabled, dtype=dtype):
                    func()
        else:
            func()



    def _print_resume_epoch_message(self, module_name,instantiate_config, resume_epoch):
        string_name = '' if module_name is None else {module_name}
        if resume_epoch == -1:
            resume_epoch_message = 'no checkpoint found, initializing from scratch'
        else:
            resume_epoch_message = f'fetching checkpoint from epoch {resume_epoch}'
        print(f'For module {string_name} with config {instantiate_config}, {resume_epoch_message}')


    def _print_skipping_loading_message(self, module_name, instantiate_config):
        string_name = '' if module_name is None else {module_name}
        print(f'Skipping loading for module {string_name} with config {instantiate_config}. This module is not a nn.Module and will not be loaded from checkpoint.')

    def _create_origin_subconfig_keys(self, module_name):
        subconfig_keys = [ModelConfigKeywords.ARCHITECTURE.value]
        if module_name is not None:
            subconfig_keys.append(module_name)
        return subconfig_keys

    def _handle_recursive_module_references(self, module_config,module_name,is_external_config=False):
        subconfig_keys = self._create_origin_subconfig_keys(module_name)        
        origin_config_name = self.data_io.fetch_config_field(
            module_config,
            subconfig_keys=subconfig_keys+[ModelConfigKeywords.ORIGIN_CONFIG.value],
            is_required=False)
        
        if origin_config_name is None:
            return module_config, module_name, is_external_config
        
        origin_module_name = self.data_io.fetch_config_field(
            module_config,
            subconfig_keys=subconfig_keys+[ModelConfigKeywords.ORIGIN_MODULE.value],
            is_required=False)

        return self._handle_recursive_module_references(origin_config_name, origin_module_name, True)

    def _can_prepare_accelerator(self):
        can_prepare = False
        for model_module in self.models:
            if isinstance(model_module, nn.Module):
                can_prepare = True
                break
        return can_prepare
    

    def _prepare_accelerator(self):
        if not self._can_prepare_accelerator():
            print('No nn.Module found, skipping accelerator preparation')
            self.is_module_wrapped_by_accelerator = [False] * len(self.models)
            return
        self.accelerator = Accelerator()

        self.is_module_wrapped_by_accelerator = []

        for i in range(len(self.models)):
            if not isinstance(self.models[i], nn.Module):
                continue
            can_set_device = issubclass(type(self.models[i]), ModelBase)
            original_model_type = type(self.models[i])
            self.models[i] = self.accelerator.prepare(self.models[i])
            accelerator_model_type = type(self.models[i])

            self.is_module_wrapped_by_accelerator.append(original_model_type != accelerator_model_type)

            if can_set_device:
                callable_model = self._get_callable_model(i)
                callable_model._set_device(self.accelerator.device)

            self.dataloader = self.accelerator.prepare(self.dataloader)
        if hasattr(self,'optimizer'):
            self.optimizer = self.accelerator.prepare(self.optimizer)
        if hasattr(self,'loss'):
            self.loss = self.accelerator.prepare(self.loss)
        if hasattr(self,'validation_dataloader'):
            self.validation_dataloader = self.accelerator.prepare(self.validation_dataloader)
        if hasattr(self,'validation_loss'):
            self.validation_loss = self.accelerator.prepare(self.validation_loss)


    def make_forward_pass(self,input_data=None):
        is_first = True
        for model_module in self.models:
            if is_first and (input_data is None):
                is_first = False
                output_data = model_module()
            else:
                output_data = model_module(input_data)
            input_data = output_data
        return output_data


    def _get_callable_model(self,index):
        model = self.models[index]
        if self.is_module_wrapped_by_accelerator[index]:
            return model.module
        return model

    def switch_model_to_train(self):
        for model_module in self.models:
            if isinstance(model_module, nn.Module):
                model_module.train()

    def switch_model_to_eval(self):
        for model_module in self.models:
            if isinstance(model_module, nn.Module):
                model_module.eval()

    def save_model_checkpoint(self,epoch):
        for model_module, module_name in zip(self.models, self.model_module_names):
            if isinstance(model_module, nn.Module):
                self.data_io.save_data(ModelCheckpoint, data_args = {
                    ModelCheckpoint.CONFIG_NAME: self.global_config_name,
                    ModelCheckpoint.EPOCH: epoch,
                    ModelCheckpoint.MODULE: module_name,
                    ModelCheckpoint.DATA: model_module.state_dict()
                })


from ml_golem.base_classes.data_io_object import DataIOObject
from ml_golem.datatypes import TrainingLog

class TensorBoardLogic(DataIOObject):
    def __init__(self,args):
        super().__init__(args)
        self.global_config_name = args.config_name


    def __call__(self):
        tensorboard_path = self.data_io.get_data_path(TrainingLog,data_args ={
            TrainingLog.CONFIG_NAME: self.global_config_name
        })
        tensorboard_command = (
            'tensorboard '
            f'--logdir {tensorboard_path}'
        )
        self.data_io.run_system_command(tensorboard_command)



from ml_golem.main_loop_logic.main_loop_messages import MainLoopMessages
from ml_golem.main_loop_logic.argparsing import get_args_and_actions



def main_loop(
        app_title = 'Default App Title',
        system_config_path = 'conf/system_configs/system_conf.yaml',
        _add_additional_args = lambda x, y: (x, y)        
        ):
    print(MainLoopMessages.BEGINNING_PROGRAM.value)
    args, actions = get_args_and_actions(app_title,system_config_path,_add_additional_args)
    for action in actions:
        if action(args):
            print(MainLoopMessages.COMPLETE_PROGRAM.value)
            return
    raise Exception(f'No action specified.')
